/* Copyright (C) 1985 Adam Fritz, 133 Main St., Afton, NY 13730 */

SGECO ( A, lda, n, IPvt, RCond, z )
   float A[][ldx] ;
   int lda, n ;
   int IPvt[] ;
   float *RCond ;
   float z[] ;

     /* Program:                                            */
     /*                                                     */
     /*    SGECO  - Factors a real matrix by Gaussian       */
     /*             elimination and computes the condition  */
     /*             RCond of the matrix.                    */
     /*                                                     */
     /* Version:                  Date:                     */
     /*                                                     */
     /*    C                         02/22/85               */
     /*    FORTRAN                   08/14/78               */
     /*                                                     */
     /* Description:                                        */
     /*                                                     */
     /*                             1                       */
     /*    RCond = ---------------------------------------- */
     /*            norm(A) * (Estimate of norm(inverse(A))) */
     /*                                                     */
     /*    where;                                           */
     /*                                                     */
     /*       Estimate = norm(z)/norm(y)                    */
     /*       A * z = y                                     */
     /*       trans(A) * y = E                              */
     /*                                                     */
     /*    and trans(A) is transpose of A. Components of E  */
     /*    are chosen to cause maximum local growth in the  */
     /*    elements of w where;                             */
     /*                                                     */
     /*       trans(U) * w = E                              */
     /*                                                     */
     /*    Vectors are rescaled to avoid overflow.          */
     /*                                                     */
     /*    If RCond is not needed, SGEFA is slightly faster.*/
     /*                                                     */
     /*    To compute:                                      */
     /*                                                     */
     /*       A * x = b       , follow SGECO by SGESL.      */
     /*       inverse(A) * c  , follow SGECO by SGESL.      */
     /*       determinant(A)  , follow SGECO by SGEDI.      */
     /*       inverse(A)      , follow SGECO by SGEDI.      */
     /*                                                     */
     /*    On entry;                                        */
     /*                                                     */
     /*       A     - Matrix to be factored.                */
     /*       lda   - Leading dimension of matrix  A.       */
     /*       n     - Order of matrix  A.                   */
     /*                                                     */
     /*    On exit;                                         */
     /*                                                     */
     /*       A     - Upper triangular matrix and  multi-   */
     /*               pliers used to obtain it. Factoriza-  */
     /*               tion can be written;                  */
     /*                  A = L * U                          */
     /*               where L  is a product of permutation  */
     /*               and unit lower triangular matrices    */
     /*               and  U  is upper triangular.          */
     /*                                                     */
     /*       IPvt  - Pivot index vector.                   */
     /*                                                     */
     /*       RCond - Estimate of reciprocal condition of A.*/
     /*               For system  A * x = B , relative per- */
     /*               turbations in  A  and  b  of size     */
     /*               Epsilon  may cause relative perturba- */
     /*               tions in  x  of size  Epsilon/RCond . */
     /*               If  RCond  is so small that the logi- */
     /*               cal expression;                       */
     /*                  (1.0 + RCond) = 1.0                */
     /*               is true, then  A  may be singular to  */
     /*               workingprecision.  In particular,     */
     /*               RCond  is zero  if exact singularity  */
     /*               is detected or the estimate under-    */
     /*               flows.                                */
     /*                                                     */
     /*       z     - Work vector whose contents are        */
     /*               usually unimportant. If A is close to */
     /*               a singular matrix, then  z  is an     */
     /*               approximate null vector in the sense  */
     /*               that;                                 */
     /*                  norm(A*z) = RCond*norm(A)*norm(z)  */
     /*                                                     */
     /* Subprograms:                                        */
     /*                                                     */
     /*    SGEFA      from LINPACK                          */
     /*    SAXPY      from BLAS                             */
     /*    SDOT       from BLAS                             */
     /*    SSCAL      from BLAS                             */
     /*    SASUM      from BLAS                             */
     /*                                                     */
     /* Authors:                                            */
     /*                                                     */
     /*    Cleve Moler                                      */
     /*    University of New Mexico                         */
     /*    Argonne National Laboratories                    */
     /*                                                     */
     /*    C and Pascal Translations;                       */
     /*                                                     */
     /*    Adam Fritz                                       */
     /*    133 Main Street                                  */
     /*    Afton, New York 13730                            */

{
   float Abs() ;
   float SASUM(), SDOT() ;
   float ek, t, wk, wkm ;
   float anorm, s, sm, ynorm ;
   int j, k, kb, kp1, l ;
                                /* Compute 1-norm of A */
   anorm = 0.0 ;
   for ( j = 0; j < n; j++ )
   {
      s = SASUM(n, &A[0][j], lda) ;
      if ( anorm < s )
         anorm = s ;
   }
                                /* Factor */
   SGEFA(A, lda, n, IPvt, &InfoCode) ;
                                /* Solve trans(U) * w = E */
   ek = 1.0 ;
   for ( j = 0; j < n; j++ )
      z[j] = 0.0 ;
   for ( k = 0; k < n; k++ )
   {
      if ( z[k] > 0.0 )
         ek = -Abs(ek) ;
      else
         ek = Abs(ek) ;
      if ( Abs(ek-z[k]) > Abs(A[k][k]) )
      {
         s = Abs(A[k][k])/Abs(ek-z[k]) ;
         SSCAL(n, s, &z[0], 1) ;
         ek = s * ek ;
      }
      wk = ek - z[k] ;
      wkm = -ek - z[k] ;
      s = Abs(wk) ;
      sm = Abs(wkm) ;
      if ( A[k][k] != 0.0 )
      {
         wk = wk/A[k][k] ;
         wkm = wkm/A[k][k] ;
      }
      else {
         wk = 1.0 ;
         wkm = 1.0 ;
      }
      kp1 = k + 1 ;
      if ( kp1 < n )
      {
         for ( j = kp1; j < n; j++ )
         {
            sm = sm + Abs(z[j] + wkm * A[k][j]) ;
            z[j] = z[j] + wk * A[k][j] ;
            s = s + Abs(z[j]) ;
         }
         if ( s < sm )
         {
            t = wkm - wk ;
            wk = wkm ;
            for ( j = kp1; j < n; j++ )
               z[j] = z[j] + t * A[k][j] ;
         }
      }
      z[k] = wk ;
   }
   s = 1.0/SASUM(n, &z[0], 1) ;
   SSCAL(n, s, &z[0], 1) ;
                                /* Solve trans(L) * y = w */
   for ( kb = 0; kb < n; kb++ )
   {
      k = n - kb - 1 ;
      if ( k < n-1 )
         z[k] = z[k] + SDOT(n-k-1, &A[k+1][k], lda, &z[k+1], 1) ;
      if ( Abs(z[k]) > 1.0 )
      {
         s = 1.0/Abs(z[k]) ;
         SSCAL(n, s, &z[0], 1) ;
      }
      l = IPvt[k] ;
      t = z[l] ;
      z[l] = z[k] ;
      z[k] = t ;
   }
   s = 1.0/SASUM(n, &z[0], 1) ;
   SSCAL(n, s, &z[0], 1) ;
                                /* Solve L * v = y */
   ynorm = 1.0 ;
   for ( k = 0; k < n; k++ )
   {
      l = IPvt[k] ;
      t = z[l] ;
      z[l] = z[k] ;
      z[k] = t ;
      if ( k < n-1 )
         SAXPY(n-k-1, t, &A[k+1][k], lda, &z[k+1], 1) ;
      if ( Abs(z[k]) > 1.0 )
      {
         s = 1.0/Abs(z[k]) ;
         SSCAL(n, s, &z[0], 1) ;
         ynorm = s * ynorm ;
      }
   }
   s = 1.0/SASUM(n, &z[0], 1) ;
   SSCAL(n, s, &z[0], 1) ;
   ynorm = s * ynorm ;
                                /* Solve  U * z = v */
   for ( kb = 0; kb < n; kb++ )
   {
      k = n - kb - 1 ;
      if ( Abs(z[k]) > Abs(A[k][k]) )
      {
         s = Abs(A[k][k])/Abs(z[k]) ;
         SSCAL(n, s, &z[0], 1) ;
         ynorm = s * ynorm ;
      }
      if ( A[k][k] != 0.0 )
         z[k] = z[k]/A[k][k] ;
      else
         z[k] = 1.0 ;
      t = -z[k] ;
      SAXPY(k, t, &A[0][k], lda, &z[0], 1) ;
   }
                                /* Set znorm = 1.0 */
   s = 1.0/SASUM(n, &z[0], 1) ;
   SSCAL(n, s, &z[0], 1) ;
   ynorm = s * ynorm ;
   if ( anorm != 0.0 )
      *RCond = ynorm/anorm ;
   else
      *RCond = 0.0 ;
}

/* Copyright (C) 1985 Adam Fritz, 133 Main St., Afton, NY 13730 */
